import glob
import logging
import os
import shutil
import six
from six import string_types
from mdt.components_loader import BatchProfilesLoader, get_model
from mdt.data_loaders.protocol import ProtocolLoader
from mdt.masking import create_write_median_otsu_brain_mask
from mdt.models.cascade import DMRICascadeModelInterface
from mdt.protocols import load_protocol, auto_load_protocol
from mdt.utils import split_image_path, AutoDict, load_problem_data
from mdt.nifti import load_nifti

__author__ = 'Robbert Harms'
__date__ = "2015-08-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BatchProfile(object):

    def __init__(self):
        """Batch profiles encapsulate information about the subjects and modelling settings.

        Suppose you have a directory full of subjects that you want to analyze with a few models. One way to do that
        is to write some scripts yourself that walk through the directory and fit the models to the subjects. The other
        way would be to implement a :class:`BatchProfile` that contains details about your directory structure and let
        :func:`mdt.batch_fit` fetch all the subjects for you.

        Batch profiles contain a list with subject information (see :class:`SubjectInfo`) and a list of models
        we wish to apply to these subjects. Furthermore each profile should support some functionality that checks
        if this profile is suitable for a given directory. Using those functions the :func:`mdt.batch_fit` can try
        to auto-recognize the batch profile to use based on the profile that is suitable and returns the most subjects.
        """
        self._root_dir = ''

    def set_root_dir(self, root_dir):
        """Set the root dir. That is, the directory we search for subjects.

        This is function level dependency injection.

        Args:
            root_dir (str): the root dir to use
        """
        self._root_dir = root_dir

    def get_root_dir(self):
        """Get the root dir this profile uses.

        Returns:
            str: the root dir this batch profile uses.
        """
        return self._root_dir

    def get_models_to_fit(self):
        """Get the list of models we want to fit to every found subject.

        The models can either be real model objects, or strings with the model names.

        Returns:
            :class:`list`: the list of models we want to fit to the subjects
        """

    def get_subjects(self):
        """Get the information about all the subjects in the current folder.

        Returns:
            list of :class`SubjectInfo`: the information about the found subjects
        """

    def profile_suitable(self):
        """Check if this directory can be used to use subjects from using this batch fitting profile.

        This is used for auto detecting the best batch fitting profile to use for loading
        subjects from the given root dir.

        Returns:
            boolean: true if this batch fitting profile can use the subjects in the current root directory,
                false otherwise.
        """

    def get_subjects_count(self):
        """Get the number of subjects this batch fitting profile can use from the current root directory.

        Returns:
            int: the number of subjects this batch fitting profile can use from the current root directory.
        """


class SimpleBatchProfile(BatchProfile):

    def __init__(self):
        """A base class for quickly implementing a batch profile.

        Implementing classes need only implement the method :meth:`_get_subjects`, then this class will handle the rest.
        """
        super(SimpleBatchProfile, self).__init__()
        self._subjects_found = None
        self._output_base_dir = 'output'
        self._output_sub_dir = None
        self._append_mask_name_to_output_sub_dir = True
        self.models_to_fit = ('BallStick_r3 (Cascade)',
                              'Tensor (Cascade)',
                              'NODDI (Cascade)',
                              'CHARMED_r1 (Cascade)',
                              'CHARMED_r2 (Cascade)',
                              'CHARMED_r3 (Cascade)')

    @property
    def output_base_dir(self):
        return self._output_base_dir

    @output_base_dir.setter
    def output_base_dir(self, output_base_dir):
        self._output_base_dir = output_base_dir
        self._subjects_found = None

    @property
    def append_mask_name_to_output_sub_dir(self):
        return self._append_mask_name_to_output_sub_dir

    @append_mask_name_to_output_sub_dir.setter
    def append_mask_name_to_output_sub_dir(self, append_mask_name_to_output_sub_dir):
        self._append_mask_name_to_output_sub_dir = append_mask_name_to_output_sub_dir

    @property
    def output_sub_dir(self):
        return self._output_sub_dir

    @output_sub_dir.setter
    def output_sub_dir(self, output_sub_dir):
        self._output_sub_dir = output_sub_dir
        self._subjects_found = None

    def get_models_to_fit(self):
        return self.models_to_fit

    def get_subjects(self):
        if not self._subjects_found:
            self._subjects_found = self._get_subjects()

        return self._subjects_found

    def profile_suitable(self):
        if not self._subjects_found:
            self._subjects_found = self._get_subjects()

        return len(self._subjects_found) > 0

    def get_subjects_count(self):
        if not self._subjects_found:
            self._subjects_found = self._get_subjects()

        return len(self._subjects_found)

    def _autoload_noise_std(self, subject_id, file_path=None):
        """Try to autoload the noise standard deviation from a noise_std file.

        Args:
            subject_id (str): the subject for which to use the noise std.
            file_path (str): optionally provide the exact file to use.

        Returns:
            float or None: a float if a float could be loaded from a file noise_std, else nothing.
        """
        file_path = file_path or os.path.join(self._root_dir, subject_id, 'noise_std')
        noise_std_files = glob.glob(file_path + '*')
        if len(noise_std_files):
            with open(noise_std_files[0], 'r') as f:
                return float(f.read())
        return None

    def _get_subjects(self):
        """Get the matching subjects from the given root dir.

        This is the only function that should be implemented to get up and running.

        Returns:
            list of SubjectInfo: the information about the found subjects
        """
        return []

    def _get_subject_output_dir(self, subject_id, mask_fname, subject_base_dir=None):
        """Helper function for generating the output directory for a subject.

        Args:
            subject_id (str): the id of the subject to use
            mask_fname (str): the name of the mask we are using for this subject
            subject_base_dir (str): the base directory for this subject, defaults to
                self._root_dir / subject_id / self.output_base_dir

        Returns:
            str: the path for the output directory
        """
        output_dir = subject_base_dir or os.path.join(self._root_dir, subject_id, self.output_base_dir)

        if self.output_sub_dir:
            output_dir = os.path.join(output_dir, self.output_sub_dir)

        if self._append_mask_name_to_output_sub_dir and mask_fname:
            output_dir = os.path.join(output_dir, split_image_path(mask_fname)[1])

        return output_dir

    def _get_first_existing_file(self, filenames, default=None, prepend_path=None):
        """Tries a list of filenames and returns the first filename in the list that exists.

        Args:
            filenames (iterator): the list of filenames to search for existence
            default (str): the default value returned if none of the filenames existed
            prepend_path (str): the path to optionally prepend to every file before checking existence

        Returns:
            str: the filename of the first existing file, if prepend path is set it is included.
        """
        for fname in filenames:
            if fname:
                if prepend_path:
                    fname = os.path.join(prepend_path, fname)
                if os.path.isfile(fname):
                    return fname
        return default

    def _get_first_existing_nifti(self, filenames, default=None, prepend_path=None):
        """Tries a list of filenames and returns the first filename in the list that exists.

        Additional to the method :meth:`_get_first_existing_file`, this additionally tries to see for every filename
        if a file with extension '.nii' or with '.nii.gz' exists (in that order) for that filename. If so,
        the path with the added extension is returned.

        Args:
            filenames (iterator): the list of filenames to search for existence, does additional extension lookup
                per filename
            default (str): the default value returned if none of the filenames existed
            prepend_path (str): the path to optionally prepend to every file before checking existence

        Returns:
            str: the filename of the first existing file, can contain an extra extension for the returned filename.
        """
        for fname in filenames:
            resolve_extension = self._get_first_existing_file([fname, fname + '.nii', fname + '.nii.gz'],
                                                              prepend_path=prepend_path)
            if resolve_extension:
                return resolve_extension
        return default

    def _autoload_protocol(self, path, protocols_to_try=(), bvecs_to_try=(), bvals_to_try=(), protocol_columns=None):
        prtcl_fname = self._get_first_existing_file(protocols_to_try, prepend_path=path)
        bval_fname = self._get_first_existing_file(bvals_to_try, prepend_path=path)
        bvec_fname = self._get_first_existing_file(bvecs_to_try, prepend_path=path)

        return BatchFitProtocolLoader(
            path, protocol_fname=prtcl_fname, bvec_fname=bvec_fname,
            bval_fname=bval_fname, protocol_columns=protocol_columns)


class SubjectInfo(object):

    @property
    def subject_id(self):
        """Get the ID of this subject.

        Returns:
            str: the id of this subject
        """
        return ''

    @property
    def output_dir(self):
        """Get the output folder for this subject.

        Returns:
            str: the output folder
        """
        return ''

    def get_problem_data(self):
        """Get the DMRIProblemData for this subject.

        This is the data we will use during model fitting.

        Returns:
            :class:`~mdt.utils.DMRIProblemData`: the problem data to use during model fitting
        """

    def get_mask_filename(self):
        """Get the filename of the mask to use.

        Returns:
            str: the filename of the mask to use
        """


class SimpleSubjectInfo(SubjectInfo):

    def __init__(self, subject_id, dwi_fname, protocol_loader, mask_fname, output_dir, gradient_deviations=None,
                 use_gradient_deviations=True, noise_std=None):
        """This class contains all the information about found subjects during batch fitting.

        It is returned by the method get_subjects() from the class BatchProfile.

        Args:
            subject_id (str): the subject id
            dwi_fname (str): the filename with path to the dwi image
            protocol_loader (ProtocolLoader): the protocol loader that can use us the protocol
            mask_fname (str): the filename of the mask to use. If None a mask is auto generated.
            output_dir (str): the output directory
            gradient_deviations (str) if given, the path to the gradient deviations
            use_gradient_deviations (boolean): if we use the gradient deviations or not
            noise_std (float, ndarray, str): either None for automatic noise detection or a float with the noise STD
                to use during fitting or an ndarray with one value per voxel.
        """
        self._subject_id = subject_id
        self._dwi_fname = dwi_fname
        self._protocol_loader = protocol_loader
        self._mask_fname = mask_fname
        self._output_dir = output_dir
        self._gradient_deviations = gradient_deviations
        self._use_gradient_deviations = use_gradient_deviations
        self._noise_std = noise_std

        if self._mask_fname is None:
            self._mask_fname = os.path.join(self.output_dir, 'auto_generated_mask.nii.gz')

    @property
    def subject_id(self):
        return self._subject_id

    @property
    def output_dir(self):
        return self._output_dir

    def get_problem_data(self):
        protocol = self._protocol_loader.get_protocol()
        brain_mask_fname = self.get_mask_filename()
        return load_problem_data(self._dwi_fname, protocol, brain_mask_fname,
                                 gradient_deviations=self._get_gradient_deviations(), noise_std=self._noise_std)

    def get_subject_id(self):
        return self.subject_id

    def get_mask_filename(self):
        if not os.path.isfile(self._mask_fname):
            logger = logging.getLogger(__name__)
            logger.info('Creating a brain mask for subject {0}'.format(self.subject_id))

            protocol = self._protocol_loader.get_protocol()
            create_write_median_otsu_brain_mask(self._dwi_fname, protocol, self._mask_fname)

        return self._mask_fname

    def _get_gradient_deviations(self):
        if self._use_gradient_deviations and self._gradient_deviations is not None:
            return load_nifti(self._gradient_deviations).get_data()
        return None


class BatchSubjectSelection(object):

    def get_selection(self, subjects):
        """Get the selection of subjects from the given list of subjects.

        Args:
            subjects (list of :class:`SubjectInfo`): the list of subjects from which we can choose which one to process

        Returns:
            list of :class:`SubjectInfo`: the given list or a subset of the given list with the subjects to process.
        """
        pass


class AllSubjects(BatchSubjectSelection):

    def __init__(self):
        """Selects all subjects for use in the processing"""
        super(AllSubjects, self).__init__()

    def get_selection(self, subjects):
        return subjects


class SelectedSubjects(BatchSubjectSelection):

    def __init__(self, subject_ids=None, indices=None, start_from=None):
        """Only process the selected subjects.

        This method allows either a selection by index (unsafe for the order may change) or by subject name/ID (more
        safe in general). If ``start_from`` is given it additionally limits the list of selected subjects to include
        only those after that index.

        This essentially creates three different subsets of the given list of subjects and it will only process
        the subjects in the intersection of all those sets.

        Set any one of the options to None to ignore that option.

        Args:
            subject_ids (list of str): the list of names of subjects to process
            indices (list/tuple of int): the list of indices of subjects we wish to process
            start_from (list or int): the index of the name of the subject from which we want to start processing.
        """
        self.subject_ids = subject_ids
        self.indices = indices
        self.start_from = start_from

    def get_selection(self, subjects):
        starting_pos = self._get_starting_pos(subjects)

        if self.indices is None and self.subject_ids is None:
            return subjects[starting_pos:]

        if self.indices:
            subjects = [subject for ind, subject in enumerate(subjects) if ind in self.indices and ind >= starting_pos]

        if self.subject_ids:
            subjects = list(filter(lambda subject: subject.subject_id in self.subject_ids, subjects))

        return subjects

    def _get_starting_pos(self, subjects):
        if self.start_from is None:
            return 0

        if isinstance(self.start_from, six.string_types):
            for ind, subject in enumerate(subjects):
                if subject.subject_id == self.start_from:
                    return ind

        for ind, subject in enumerate(subjects):
            if ind == int(self.start_from):
                return ind


class BatchFitProtocolLoader(ProtocolLoader):

    def __init__(self, base_dir, protocol_fname=None, protocol_columns=None, bvec_fname=None, bval_fname=None):
        """A simple protocol loader for loading a protocol from a protocol file or bvec/bval files.

        This either loads the protocol file if present, or autoloads the protocol using the auto_load_protocol
        from the protocol module.
        """
        super(BatchFitProtocolLoader, self).__init__()
        self._base_dir = base_dir
        self._protocol_fname = protocol_fname
        self._bvec_fname = bvec_fname
        self._bval_fname = bval_fname
        self._protocol_columns = protocol_columns

    def get_protocol(self):
        super(BatchFitProtocolLoader, self).get_protocol()

        if self._protocol_fname and os.path.isfile(self._protocol_fname):
            return load_protocol(self._protocol_fname)

        return auto_load_protocol(self._base_dir, protocol_columns=self._protocol_columns,
                                  bvec_fname=self._bvec_fname, bval_fname=self._bval_fname)


class BatchFitSubjectOutputInfo(object):

    def __init__(self, subject_info, output_path, model_name):
        """This class is used in conjunction with the function :func:`run_function_on_batch_fit_output`.

        Args:
            subject_info (SubjectInfo): the information about the subject before batch fitting
            output_path (str): the full path to the directory with the maps
            model_name (str): the name of the model (not a path)
        """
        self.subject_info = subject_info
        self.output_path = output_path
        self.model_name = model_name

    @property
    def mask_name(self):
        return split_image_path(self.subject_info.get_mask_filename())[1]

    @property
    def subject_id(self):
        return self.subject_info.subject_id


class BatchFitOutputInfo(object):

    def __init__(self, data_folder, batch_profile=None, subjects_selection=None):
        """Single point of information about batch fitting results.

        Args:
            data_folder (str): The data folder with the output files
            batch_profile (:class:`BatchProfile` or str): the batch profile to use, can also be the name
                of a batch profile to use. If not given it is auto detected.
            subjects_selection (BatchSubjectSelection): the subjects to use for processing.
                If None all subjects are processed.
        """
        self._data_folder = data_folder
        self._batch_profile = batch_profile_factory(batch_profile, data_folder)
        self._subjects_selection = subjects_selection or AllSubjects()
        self._subjects = self._subjects_selection.get_selection(self._batch_profile.get_subjects())
        self._mask_paths = {}

    def subject_output_info_generator(self):
        """Generates for every subject an output info object which contains all relevant information about the subject.

        Returns:
            generator: returns an BatchFitSubjectOutputInfo per subject
        """
        model_names = self._get_composite_model_names(self._batch_profile.get_models_to_fit())

        for subject_info in self._subjects:
            for model_name in model_names:
                output_path = os.path.join(subject_info.output_dir, model_name)
                if os.path.isdir(output_path):
                    yield BatchFitSubjectOutputInfo(subject_info, output_path, model_name)

    @staticmethod
    def _get_composite_model_names(model_names):
        """Resolve the composite model names from the list of (possibly cascade) model names from the BatchProfile"""
        lookup_cache = {}

        def get_names(current_names):
            composite_model_names = []

            for model_name in current_names:
                if model_name not in lookup_cache:
                    model = get_model(model_name)
                    if isinstance(model, DMRICascadeModelInterface):
                        resolved_names = get_names(model.get_model_names())
                        lookup_cache[model_name] = resolved_names
                    else:
                        lookup_cache[model_name] = [model_name]

                composite_model_names.extend(lookup_cache[model_name])

            return composite_model_names

        return list(set(get_names(model_names)))


def run_function_on_batch_fit_output(data_folder, func, batch_profile=None, subjects_selection=None):
    """Run a function on the output of a batch fitting routine.

    This enables you to run a function on every model output from every subject. The callback python function
    should accept as single argument an instance of the class BatchFitSubjectOutputInfo.

    Args:
        data_folder (str): The data folder with the output files
        func (python function): the python function we should call for every map and model.
            This should accept as single parameter a BatchFitSubjectOutputInfo.
        batch_profile (BatchProfile class or str): the batch profile to use, can also be the name
            of a batch profile to use. If not given it is auto detected.
        subjects_selection (BatchSubjectSelection): the subjects to use for processing.
            If None all subjects are processed.

    Returns:
        dict: indexed by subject->model_name, values are the return values of the users function
    """
    output_info = BatchFitOutputInfo(data_folder, batch_profile, subjects_selection=subjects_selection)

    results = AutoDict()
    for subject in output_info.subject_output_info_generator():
        results[subject.subject_id][subject.model_name] = func(subject)

    return results.to_normal_dict()


def batch_profile_factory(batch_profile, data_folder):
    """Wrapper function for getting a batch profile.

    Args:
        batch_profile (None, string or BatchProfile): indication of the batch profile to use.
            If a string is given it is loaded from the users home folder. Else the best matching profile is returned.
        data_folder (str): the data folder we want to use the batch profile on.

    Returns:
        If the given batch profile is None we return the output from get_best_batch_profile(). If batch profile is
        a string we use it from the batch profiles loader. Else we return the input.
    """
    if batch_profile is None:
        batch_profile = get_best_batch_profile(data_folder)
    elif isinstance(batch_profile, string_types):
        batch_profile = BatchProfilesLoader().load(batch_profile)

    batch_profile.set_root_dir(data_folder)
    return batch_profile


def get_best_batch_profile(data_folder):
    """Get the batch profile that best matches the given directory.

    Args:
        data_folder (str): the directory for which to get the best batch profile.

    Returns:
        BatchProfile: the best matching batch profile.
    """
    profile_loader = BatchProfilesLoader()
    crawlers = [profile_loader.load(c) for c in profile_loader.list_all()]

    best_crawler = None
    best_subjects_count = 0
    for crawler in crawlers:
        crawler.set_root_dir(data_folder)
        if crawler.profile_suitable():
            tmp_count = crawler.get_subjects_count()
            if tmp_count > best_subjects_count:
                best_crawler = crawler
                best_subjects_count = tmp_count

    return best_crawler


def collect_batch_fit_output(data_folder, output_dir, batch_profile=None, subjects_selection=None, symlink=True,
                             symlink_absolute=False, move=False):
    """Load from the given data folder all the output files and put them into the output directory.

    The results are placed in the output folder per subject. Example: ``<output_dir>/<subject_id>/<model_name>``

    Args:
        data_folder (str): The data folder with the output files
        output_dir (str): The path to the output folder where all the files will be put.
        batch_profile (:class:`BatchProfile` or str): the batch profile to use, can also be the name
            of a batch profile to use. If not given it is auto detected.
        subjects_selection (BatchSubjectSelection): the subjects to use for processing.
            If None all subjects are processed.
        symlink (boolean): only available under Unix OS's. Creates a symlink instead of copying.
            This will create an absolute position symlink.
        symlink_absolute (boolean): if symlink is set to true, do you want an absolute symlink (True)
            or a relative one (False)
        move (boolean): instead of copying the files, move them to a new position. If set, this overrules the parameter
            symlink.
    """
    def copy_function(subject_info):
        if not os.path.exists(os.path.join(output_dir, subject_info.subject_id)):
            os.makedirs(os.path.join(output_dir, subject_info.subject_id))

        subject_out = os.path.join(output_dir, subject_info.subject_id, subject_info.model_name)

        if os.path.exists(subject_out) or os.path.islink(subject_out):
            if os.path.islink(subject_out):
                os.unlink(subject_out)
            else:
                shutil.rmtree(subject_out)

        if move:
            shutil.move(subject_info.output_path, subject_out)
        else:
            if symlink:
                if symlink_absolute:
                    os.symlink(subject_info.output_path, subject_out)
                else:
                    os.symlink(os.path.relpath(subject_info.output_path, os.path.dirname(subject_out)), subject_out)
            else:
                shutil.copytree(subject_info.output_path, subject_out)

    run_function_on_batch_fit_output(data_folder, copy_function, batch_profile=batch_profile,
                                     subjects_selection=subjects_selection)



from mdt.components_loader import bind_function
from mdt.models.compartments import CompartmentConfig
from mot.model_building.cl_functions.library_functions import CerfDawson

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDI_EC(CompartmentConfig):

    parameter_list = ('g', 'b', 'd', 'dperp0', 'theta', 'phi', 'kappa')
    dependency_list = (CerfDawson(),)

    @bind_function
    def get_extra_results_maps(self, results_dict):
        return self._get_vector_result_maps(results_dict[self.name + '.theta'],
                                            results_dict[self.name + '.phi'])



from mdt.processing_strategies import ChunksProcessingStrategy

__author__ = 'Robbert Harms'
__date__ = "2015-11-29"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


meta_info = {'title': 'Fit in chunks of voxel ranges',
             'description': 'Processes a model in chunks defined by ranges of voxels.'}


class VoxelRange(ChunksProcessingStrategy):

    def __init__(self, nmr_voxels=40000, **kwargs):
        """Optimize a given dataset in batches of the given number of voxels

        Args:
            nmr_voxels (int): the number of voxels per batch

        Attributes:
            nmr_voxels (int): the number of voxels per chunk
        """
        super(VoxelRange, self).__init__(**kwargs)
        self.nmr_voxels = nmr_voxels

    def _chunks_generator(self, model, problem_data, output_path, worker, total_roi_indices):
        for ind_start in range(0, len(total_roi_indices), self.nmr_voxels):
            ind_end = min(len(total_roi_indices), ind_start + self.nmr_voxels)
            yield total_roi_indices[ind_start:ind_end]



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'main_gui.ui'
#
# Created by: PyQt5 UI code generator 5.4.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(870, 650)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/main_gui/logo"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MainWindow.setWindowIcon(icon)
        MainWindow.setTabShape(QtWidgets.QTabWidget.Rounded)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setContentsMargins(3, 0, 3, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.splitter = QtWidgets.QSplitter(self.centralwidget)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.MainTabs = QtWidgets.QTabWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.MainTabs.sizePolicy().hasHeightForWidth())
        self.MainTabs.setSizePolicy(sizePolicy)
        self.MainTabs.setObjectName("MainTabs")
        self.fitModelTab = QtWidgets.QWidget()
        self.fitModelTab.setObjectName("fitModelTab")
        self.MainTabs.addTab(self.fitModelTab, "")
        self.generateBrainMaskTab = QtWidgets.QWidget()
        self.generateBrainMaskTab.setObjectName("generateBrainMaskTab")
        self.MainTabs.addTab(self.generateBrainMaskTab, "")
        self.generateROIMaskTab = QtWidgets.QWidget()
        self.generateROIMaskTab.setObjectName("generateROIMaskTab")
        self.MainTabs.addTab(self.generateROIMaskTab, "")
        self.generateProtocolTab = QtWidgets.QWidget()
        self.generateProtocolTab.setObjectName("generateProtocolTab")
        self.MainTabs.addTab(self.generateProtocolTab, "")
        self.viewResultsTab = QtWidgets.QWidget()
        self.viewResultsTab.setObjectName("viewResultsTab")
        self.MainTabs.addTab(self.viewResultsTab, "")
        self.loggingTextBox = QtWidgets.QPlainTextEdit(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.loggingTextBox.sizePolicy().hasHeightForWidth())
        self.loggingTextBox.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Droid Sans Mono")
        self.loggingTextBox.setFont(font)
        self.loggingTextBox.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.loggingTextBox.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.loggingTextBox.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.loggingTextBox.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.loggingTextBox.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.loggingTextBox.setReadOnly(True)
        self.loggingTextBox.setPlainText("")
        self.loggingTextBox.setTabStopWidth(80)
        self.loggingTextBox.setObjectName("loggingTextBox")
        self.verticalLayout_2.addWidget(self.splitter)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, 0, 8, -1)
        self.horizontalLayout.setSpacing(8)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.executionStatusLabel = QtWidgets.QLabel(self.centralwidget)
        self.executionStatusLabel.setObjectName("executionStatusLabel")
        self.horizontalLayout.addWidget(self.executionStatusLabel)
        self.executionStatusIcon = QtWidgets.QLabel(self.centralwidget)
        self.executionStatusIcon.setObjectName("executionStatusIcon")
        self.horizontalLayout.addWidget(self.executionStatusIcon)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 870, 27))
        self.menubar.setObjectName("menubar")
        self.menuMenu = QtWidgets.QMenu(self.menubar)
        self.menuMenu.setObjectName("menuMenu")
        self.menuHelp = QtWidgets.QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        MainWindow.setMenuBar(self.menubar)
        self.actionExit = QtWidgets.QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.action_RuntimeSettings = QtWidgets.QAction(MainWindow)
        self.action_RuntimeSettings.setObjectName("action_RuntimeSettings")
        self.actionAbout = QtWidgets.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.menuMenu.addAction(self.action_RuntimeSettings)
        self.menuMenu.addSeparator()
        self.menuMenu.addAction(self.actionExit)
        self.menuHelp.addAction(self.actionAbout)
        self.menubar.addAction(self.menuMenu.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        self.MainTabs.setCurrentIndex(0)
        self.actionExit.triggered.connect(MainWindow.close)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Maastricht Diffusion Toolbox"))
        self.MainTabs.setTabText(self.MainTabs.indexOf(self.fitModelTab), _translate("MainWindow", "Fit model"))
        self.MainTabs.setTabText(self.MainTabs.indexOf(self.generateBrainMaskTab), _translate("MainWindow", "Generate brain mask"))
        self.MainTabs.setTabText(self.MainTabs.indexOf(self.generateROIMaskTab), _translate("MainWindow", "Generate ROI mask"))
        self.MainTabs.setTabText(self.MainTabs.indexOf(self.generateProtocolTab), _translate("MainWindow", "Generate protocol file"))
        self.MainTabs.setTabText(self.MainTabs.indexOf(self.viewResultsTab), _translate("MainWindow", "View results"))
        self.executionStatusLabel.setText(_translate("MainWindow", "TextLabel"))
        self.executionStatusIcon.setText(_translate("MainWindow", "TextLabel"))
        self.menuMenu.setTitle(_translate("MainWindow", "&File"))
        self.menuHelp.setTitle(_translate("MainWindow", "&Help"))
        self.actionExit.setText(_translate("MainWindow", "&Quit"))
        self.actionExit.setShortcut(_translate("MainWindow", "Ctrl+Q"))
        self.action_RuntimeSettings.setText(_translate("MainWindow", "&Runtime settings"))
        self.actionAbout.setText(_translate("MainWindow", "&About"))

from . import main_gui_rc



from mdt.components_loader import load_component
from mdt.processing_strategies import SimpleProcessingStrategy

__author__ = 'Robbert Harms'
__date__ = "2015-11-29"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


meta_info = {'title': 'Applies the VoxelRange strategy depending on the protocol.',
             'description': 'This looks at the size of the protocol and based on that determines the voxel range.'}


class ProtocolDependent(SimpleProcessingStrategy):

    def __init__(self, steps=((0, None), (100, 50000), (200, 20000)), **kwargs):
        """A meta strategy using VoxelRange AllVoxelsAtOnce depending on the protocol length

        This will look at the protocol of the given model and determine, based on the number of rows in the protocol,
        which voxel range to use.

        A voxel range of None or 0 means we want to fit all the voxels at once (this will use the AllVoxelsAtOnce
        strategy for that).

        During lookup of the protocol length we take the maximum step that is lower than the protocol length. If no
        suitable lookup is present, we use the AllVoxelsAtOnce strategy.

        Args:
            steps (list[tuple[int, int]]): the steps of the voxel ranges. The first item in the tuple is the
                protocol length, the second the voxel range. We assume that voxel ranges are in ascending order.
        """
        super(ProtocolDependent, self).__init__(**kwargs)
        self._steps = steps

    def run(self, model, problem_data, output_path, recalculate, worker_generator):
        strategy = self._get_strategy(problem_data)
        return strategy.run(model, problem_data, output_path, recalculate, worker_generator)

    def _get_strategy(self, problem_data):
        for col_length, voxel_range in reversed(self._steps):
            if int(col_length) < problem_data.get_nmr_inst_per_problem():
                if voxel_range:
                    return load_component('processing_strategies', 'VoxelRange', nmr_voxels=int(voxel_range),
                                          tmp_dir=self._tmp_dir, honor_voxels_to_analyze=self._honor_voxels_to_analyze)

        return load_component('processing_strategies', 'AllVoxelsAtOnce', tmp_dir=self._tmp_dir,
                              honor_voxels_to_analyze=self._honor_voxels_to_analyze)



import pyopencl as cl
import numpy as np
from mot.utils import get_float_type_def
from mot.cl_routines.base import CLRoutine
from mot.load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CalculateEigenvectors(CLRoutine):

    def convert_theta_phi_psi(self, theta_roi, phi_roi, psi_roi, double_precision=False):
        """Calculate the eigenvectors from the given theta, phi and psi angles.

        This will return the eigenvectors unsorted (since we know nothing about the eigenvalues).

        Args:
            theta_roi (ndarray): The list of theta's per voxel in the ROI
            phi_roi (ndarray): The list of phi's per voxel in the ROI
            psi_roi (ndarray): The list of psi's per voxel in the ROI
            double_precision (boolean): if we want to use float (set it to False) or double (set it to True)

        Returns:
            The three eigenvectors per voxel in the ROI. The return matrix is of shape (n, 3, 3) where n is the number
            of voxels and the second dimension holds the number of vectors and the last dimension the direction
            per vector. In other words, this gives for one voxel the matrix::

                [evec_1_x, evec_1_y, evec_1_z,
                 evec_2_x, evec_2_y, evec_2_z
                 evec_3_x, evec_3_y, evec_3_z]

            The resulting eigenvectors are the same as those from the Tensor compartment model.
        """
        np_dtype = np.float32
        if double_precision:
            np_dtype = np.float64

        theta_roi = np.require(theta_roi, np_dtype, requirements=['C', 'A', 'O'])
        phi_roi = np.require(phi_roi, np_dtype, requirements=['C', 'A', 'O'])
        psi_roi = np.require(psi_roi, np_dtype, requirements=['C', 'A', 'O'])

        rows = theta_roi.shape[0]
        evecs = np.zeros((rows, 3, 3), dtype=np_dtype, order='C')

        workers = self._create_workers(lambda cl_environment: _CEWorker(cl_environment, self.get_compile_flags_list(),
                                                                        theta_roi, phi_roi,
                                                                        psi_roi, evecs, double_precision))
        self.load_balancer.process(workers, rows)
        return evecs


class _CEWorker(Worker):

    def __init__(self, cl_environment, compile_flags, theta_roi, phi_roi, psi_roi, evecs, double_precision):
        super(_CEWorker, self).__init__(cl_environment)

        self._theta_roi = theta_roi
        self._phi_roi = phi_roi
        self._psi_roi = psi_roi
        self._evecs = evecs
        self._double_precision = double_precision
        self._all_buffers, self._evecs_buf = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def __del__(self):
        list(buffer.release() for buffer in self._all_buffers)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        event = self._kernel.generate_tensor(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                             global_offset=(int(range_start),))
        return [self._enqueue_readout(self._evecs_buf, self._evecs, range_start, range_end, [event])]

    def _create_buffers(self):
        thetas_buf = cl.Buffer(self._cl_run_context.context,
                               cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                               hostbuf=self._theta_roi)
        phis_buf = cl.Buffer(self._cl_run_context.context,
                             cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                             hostbuf=self._phi_roi)
        psis_buf = cl.Buffer(self._cl_run_context.context,
                             cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                             hostbuf=self._psi_roi)
        evecs_buf = cl.Buffer(self._cl_run_context.context,
                              cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                              hostbuf=self._evecs)

        all_buffers = [thetas_buf, phis_buf, psis_buf, evecs_buf]
        return all_buffers, evecs_buf

    def _get_kernel_source(self):
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += '''
            __kernel void generate_tensor(
                global mot_float_type* thetas,
                global mot_float_type* phis,
                global mot_float_type* psis,
                global mot_float_type* evecs
                ){
                    int gid = get_global_id(0);

                    mot_float_type theta = thetas[gid];
                    mot_float_type phi = phis[gid];
                    mot_float_type psi = psis[gid];

                    mot_float_type cos_theta;
                    mot_float_type sin_theta = sincos(theta, &cos_theta);
                    mot_float_type cos_phi;
                    mot_float_type sin_phi = sincos(phi, &cos_phi);
                    mot_float_type cos_psi;
                    mot_float_type sin_psi = sincos(psi, &cos_psi);

                    mot_float_type4 n1 = (mot_float_type4)(cos_phi * sin_theta, sin_phi * sin_theta, cos_theta, 0.0);

                    // rotate around n1
                    mot_float_type tmp = sin(theta+(M_PI_2)); // using tmp as the rotation factor (90 degrees)
                    mot_float_type4 n2 = (mot_float_type4)(tmp * cos_phi, tmp * sin_phi, cos(theta+(M_PI_2)), 0.0);
                    tmp = select(1, -1, n1.z < 0 || ((n1.z == 0.0) && n1.x < 0.0)); // using tmp as the multiplier
                    n2 = n2 * cos_psi + (cross(n2, tmp * n1) * sin_psi) + (tmp * n1 * dot(tmp * n1, n2) * (1-cos_psi));

                    mot_float_type4 n3 = cross(n1, n2);

                    evecs[gid*9] = n1.x;
                    evecs[gid*9 + 1] = n1.y;
                    evecs[gid*9 + 2] = n1.z;

                    evecs[gid*9 + 3] = n2.x;
                    evecs[gid*9 + 4] = n2.y;
                    evecs[gid*9 + 5] = n2.z;

                    evecs[gid*9 + 6] = n3.x;
                    evecs[gid*9 + 7] = n3.y;
                    evecs[gid*9 + 8] = n3.z;
            }
        '''
        return kernel_source



from mdt.models.compartments import CompartmentConfig

__author__ = 'Francisco J. Fritz'
__date__ = "2016-09-06"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ExpT1DecTM(CompartmentConfig):

    parameter_list = ('SEf', 'TM', 'flip_angle', 'Refoc_fa1', 'Refoc_fa2', 'T1')
    cl_code = 'return pow(0.5, SEf) * sin(flip_angle) * sin(Refoc_fa1) * sin(Refoc_fa2) * exp(-TM / T1);'



from mdt.components_loader import bind_function
from mdt.models.compartments import CompartmentConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Zeppelin(CompartmentConfig):

    parameter_list = ('g', 'b', 'd', 'dperp0', 'theta', 'phi')
    cl_code = '''
        return exp(-b *
                    (((d - dperp) *
                          pown(dot(g, (mot_float_type4)(cos(phi) * sin(theta),
                                                        sin(phi) * sin(theta), cos(theta), 0.0)), 2)
                    ) + dperp));
    '''

    @bind_function
    def get_extra_results_maps(self, results_dict):
        return self._get_vector_result_maps(results_dict[self.name + '.theta'],
                                            results_dict[self.name + '.phi'])



import matplotlib
import signal
import yaml
from PyQt5.QtCore import QObject, pyqtSlot
from PyQt5.QtCore import QTimer
from PyQt5.QtCore import QUrl
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QDialogButtonBox
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QMainWindow

from mdt.gui.maps_visualizer.actions import NewConfigAction, SetMapsToShow
from mdt.gui.maps_visualizer.config_tabs.tab_general import TabGeneral
from mdt.gui.maps_visualizer.config_tabs.tab_map_specific import TabMapSpecific
from mdt.gui.maps_visualizer.config_tabs.tab_textual import TabTextual
from mdt.gui.maps_visualizer.design.ui_save_image_dialog import Ui_SaveImageDialog

matplotlib.use('Qt5Agg')

import mdt
from mdt.gui.maps_visualizer.base import Controller, PlottingFrameInfoViewer
from mdt.visualization.maps.base import DataInfo, MapPlotConfig
from mdt.gui.maps_visualizer.renderers.matplotlib_renderer import MatplotlibPlotting
from mdt.gui.model_fit.design.ui_about_dialog import Ui_AboutDialog
from mdt.gui.utils import center_window, DirectoryImageWatcher, QtManager
from mdt.gui.maps_visualizer.design.ui_MainWindow import Ui_MapsVisualizer


class MapsVisualizerWindow(QMainWindow, Ui_MapsVisualizer):

    def __init__(self, controller, parent=None, enable_directory_watcher=True):
        super(MapsVisualizerWindow, self).__init__(parent)
        self.setupUi(self)

        self._controller = controller
        self._controller.new_data.connect(self.set_new_data)
        self._controller.new_config.connect(self.set_new_config)

        self._directory_watcher = DirectoryImageWatcher()
        if enable_directory_watcher:
            self._directory_watcher.image_updates.connect(self._directory_changed)

        self._coordinates_label = QLabel()

        self.statusBar().addPermanentWidget(self._coordinates_label)
        self.statusBar().setStyleSheet("QStatusBar::item { border: 0px solid black }; ")

        self.plotting_info_to_statusbar = PlottingFrameInfoToStatusBar(self._coordinates_label)
        self.plotting_frame = MatplotlibPlotting(controller, parent=parent,
                                                 plotting_info_viewer=self.plotting_info_to_statusbar)
        self.plotLayout.addWidget(self.plotting_frame)

        self.tab_general = TabGeneral(controller, self)
        self.generalTabPosition.addWidget(self.tab_general)

        self.tab_specific = TabMapSpecific(controller, self)
        self.mapSpecificTabPosition.addWidget(self.tab_specific)

        self.tab_textual = TabTextual(controller, self)
        self.textInfoTabPosition.addWidget(self.tab_textual)

        self.auto_rendering.setChecked(True)
        self.auto_rendering.stateChanged.connect(self._set_auto_rendering)
        self.manual_render.clicked.connect(lambda: self.plotting_frame.redraw())

        self.actionAbout.triggered.connect(lambda: AboutDialog(self).exec_())
        self.actionOpen_directory.triggered.connect(self._open_new_directory)
        self.actionSaveImage.triggered.connect(lambda: ExportImageDialog(self, self.plotting_frame).exec_())
        self.actionBrowse_to_current_folder.triggered.connect(
            lambda: QDesktopServices.openUrl(QUrl.fromLocalFile(self._controller.get_data().directory)))
        self.actionSave_settings.triggered.connect(lambda: self._save_settings())
        self.actionLoad_settings.triggered.connect(lambda: self._load_settings())

        self.undo_config.setDisabled(not self._controller.has_undo())
        self.redo_config.setDisabled(not self._controller.has_redo())

        self.undo_config.clicked.connect(lambda: self._controller.undo())
        self.redo_config.clicked.connect(lambda: self._controller.redo())

    @pyqtSlot(DataInfo)
    def set_new_data(self, data_info):
        self.actionBrowse_to_current_folder.setDisabled(self._controller.get_data().directory is None)

        if self._controller.get_data().directory is not None:
            self._directory_watcher.set_directory(self._controller.get_data().directory)

    @pyqtSlot(MapPlotConfig)
    def set_new_config(self, config):
        self.undo_config.setDisabled(not self._controller.has_undo())
        self.redo_config.setDisabled(not self._controller.has_redo())

    def _open_new_directory(self):
        initial_dir = self._controller.get_data().directory
        new_dir = QFileDialog(self).getExistingDirectory(caption='Select a folder', directory=initial_dir)
        if new_dir:
            data = DataInfo.from_dir(new_dir)
            config = MapPlotConfig()

            if len(data.maps):
                config.slice_index = data.get_max_slice_index(config.dimension) // 2

            if self._controller.get_data().maps:
                start_gui(data, config, app_exec=False)
            else:
                self._controller.set_data(data, config)

    @pyqtSlot(tuple, tuple)
    def _directory_changed(self, additions, removals):
        data = DataInfo.from_dir(self._controller.get_data().directory)
        config = self._controller.get_config()
        config.maps_to_show = [m for m in config.maps_to_show if m not in removals]
        self._controller.set_data(data, config)

    @pyqtSlot()
    def _set_auto_rendering(self):
        auto_render = self.auto_rendering.isChecked()
        self.plotting_frame.set_auto_rendering(auto_render)
        if auto_render:
            self.plotting_frame.redraw()

    def send_sigint(self, *args):
        self.close()

    def _save_settings(self):
        """Save the current settings as a text file.

        Args:
            file_name: the filename to write to
        """
        config_file = ['conf (*.conf)', 'All files (*)']
        file_name, used_filter = QFileDialog().getSaveFileName(caption='Select the GUI config file',
                                                               filter=';;'.join(config_file))
        if file_name:
            with open(file_name, 'w') as f:
                f.write(self._controller.get_config().to_yaml())

    def _load_settings(self):
        config_file = ['conf (*.conf)', 'All files (*)']
        file_name, used_filter = QFileDialog().getOpenFileName(caption='Select the GUI config file',
                                                               filter=';;'.join(config_file))
        if file_name:
            with open(file_name, 'r') as f:
                try:
                    self._controller.apply_action(NewConfigAction(MapPlotConfig.from_yaml(f.read())))
                except yaml.parser.ParserError:
                    pass
                except yaml.scanner.ScannerError:
                    pass
                except ValueError:
                    pass

    def set_window_title(self, title):
        self.setWindowTitle('MDT Maps Visualizer - {}'.format(title))


class PlottingFrameInfoToStatusBar(PlottingFrameInfoViewer):

    def __init__(self, status_bar_label):
        super(PlottingFrameInfoToStatusBar, self).__init__()
        self._status_bar_label = status_bar_label

    def set_voxel_info(self, onscreen_coords, data_index, value):
        super(PlottingFrameInfoToStatusBar, self).set_voxel_info(onscreen_coords, data_index, value)
        self._status_bar_label.setText("{}, {}, {:.3f}".format(onscreen_coords, data_index, value))

    def clear_voxel_info(self):
        super(PlottingFrameInfoToStatusBar, self).clear_voxel_info()
        self._status_bar_label.setText("")


class ExportImageDialog(Ui_SaveImageDialog, QDialog):

    previous_values = {'width': None, 'height': None,
                       'dpi': None, 'output_file': None}

    def __init__(self, parent, plotting_frame):
        super(ExportImageDialog, self).__init__(parent)
        self.setupUi(self)
        self._plotting_frame = plotting_frame
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.buttonBox.button(QDialogButtonBox.Ok).clicked.connect(self._export_image)
        self.outputFile_box.textChanged.connect(self._update_ok_button)
        self.outputFile_chooser.clicked.connect(lambda: self._select_file())

        if self.previous_values['width']:
            self.width_box.setValue(self.previous_values['width'])
        if self.previous_values['height']:
            self.height_box.setValue(self.previous_values['height'])
        if self.previous_values['dpi']:
            self.dpi_box.setValue(self.previous_values['dpi'])
        if self.previous_values['output_file']:
            self.outputFile_box.setText(self.previous_values['output_file'])

    @pyqtSlot()
    def _update_ok_button(self):
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(self.outputFile_box.text() != '')

    def _select_file(self):
        graphical_image_filters = ['png (*.png)', 'svg (*.svg)', 'All files (*)']

        open_file, used_filter = QFileDialog().getSaveFileName(caption='Select the output file',
                                                               filter=';;'.join(graphical_image_filters))

        if open_file:
            self.outputFile_box.setText(open_file)
            self._update_ok_button()

    def _export_image(self):
        self._plotting_frame.export_image(self.outputFile_box.text(), self.width_box.value(), self.height_box.value(),
                                          dpi=self.dpi_box.value())
        self.previous_values['width'] = self.width_box.value()
        self.previous_values['height'] = self.height_box.value()
        self.previous_values['dpi'] = self.dpi_box.value()
        self.previous_values['output_file'] = self.outputFile_box.text()


class AboutDialog(Ui_AboutDialog, QDialog):

    def __init__(self, parent):
        super(AboutDialog, self).__init__(parent)
        self.setupUi(self)
        self.contentLabel.setText(self.contentLabel.text().replace('{version}', mdt.__version__))


class QtController(Controller, QObject):

    new_data = pyqtSignal(DataInfo)
    new_config = pyqtSignal(MapPlotConfig)

    def __init__(self):
        super(QtController, self).__init__()
        self._data_info = DataInfo({})
        self._actions_history = []
        self._redoable_actions = []
        self._current_config = MapPlotConfig()

    def set_data(self, data_info, config=None):
        self._data_info = data_info
        self._actions_history = []
        self._redoable_actions = []

        if not config:
            config = MapPlotConfig()
        elif not isinstance(config, MapPlotConfig):
            config = MapPlotConfig.from_dict(config.to_dict())

        config.maps_to_show = list(filter(lambda k: k in data_info.maps, config.maps_to_show))

        self._apply_config(config)
        self.new_data.emit(data_info)
        self.new_config.emit(self._current_config)

    def get_data(self):
        return self._data_info

    def set_config(self, general_config):
        applied = self._apply_config(general_config)
        if applied:
            self._actions_history.clear()
            self._redoable_actions.clear()
            self.new_config.emit(self._current_config)

    def get_config(self):
        return self._current_config

    def apply_action(self, action):
        applied = self._apply_config(action.apply(self._data_info, self._current_config))
        if applied:
            self._actions_history.append(action)
            self._redoable_actions = []
            self.new_config.emit(self._current_config)

    def undo(self):
        if len(self._actions_history):
            action = self._actions_history.pop()
            self._apply_config(action.unapply())
            self._redoable_actions.append(action)
            self.new_config.emit(self._current_config)

    def redo(self):
        if len(self._redoable_actions):
            action = self._redoable_actions.pop()
            self._apply_config(action.apply(self._data_info, self._current_config))
            self._actions_history.append(action)
            self.new_config.emit(self._current_config)

    def has_undo(self):
        return len(self._actions_history) > 0

    def has_redo(self):
        return len(self._redoable_actions) > 0

    def _apply_config(self, new_config):
        """Apply the current configuration.

        Args:
            new_config (MapPlotConfig): the new configuration to apply

        Returns:
            bool: if the configuration was applied or not. If the difference with the current configuration
                and the old one is None, False is returned. Else True is returned.
        """
        validated_config = new_config.validate(self._data_info)
        if self._current_config != validated_config:
            self._current_config = validated_config
            return True
        return False


def start_gui(data=None, config=None, controller=None, app_exec=True, show_maximized=False, window_title=None,
              enable_directory_watcher=True):
    """Start the GUI with the given data and configuration.

    Args:
        data (DataInfo): the initial set of data
        config (MapPlotConfig): the initial configuration
        controller (QtController): the controller to use in the application
        app_exec (boolean): if true we execute the Qt application, set to false to disable.
        show_maximized (true): if we want to show the window in a maximized state
        window_title (str): the title of the window
        enable_directory_watcher (boolean): if the directory watcher should be enabled/disabled.
            If the directory watcher is enabled, the viewer will automatically add new maps when added to the folder
            and also automatically remove maps when they are removed from the directory. It is useful to disable this
            if you want to have multiple viewers open with old results.

    Returns:
        MapsVisualizerWindow: the generated window
    """
    controller = controller or QtController()

    app = QtManager.get_qt_application_instance()

    # catches the sigint
    timer = QTimer()
    timer.start(500)
    timer.timeout.connect(lambda: None)

    main = MapsVisualizerWindow(controller, enable_directory_watcher=enable_directory_watcher)
    main.set_window_title(window_title)
    signal.signal(signal.SIGINT, main.send_sigint)

    center_window(main)

    if show_maximized:
        main.showMaximized()

    main.show()

    if data:
        controller.set_data(data, config)
    elif config:
        controller.set_config(config)

    QtManager.add_window(main)
    if app_exec:
        QtManager.exec_()

    return main



from mdt.visualization.maps.base import Zoom, Point
from .base import ConfigAction, SimpleConfigAction, SimpleMapSpecificConfigAction


class SetDimension(SimpleConfigAction):

    config_attribute = 'dimension'

    def _extra_actions(self, data_info, configuration):
        if self.new_value != self._previous_config.dimension:
            max_slice = data_info.get_max_slice_index(self.new_value, configuration.maps_to_show)
            if configuration.slice_index > max_slice:
                configuration = SetSliceIndex(max_slice // 2).apply(data_info, configuration)
            return SetZoom(Zoom.no_zoom()).apply(data_info, configuration)


class SetSliceIndex(SimpleConfigAction):

    config_attribute = 'slice_index'


class SetVolumeIndex(SimpleConfigAction):

    config_attribute = 'volume_index'


class SetMapsToShow(SimpleConfigAction):

    config_attribute = 'maps_to_show'


class SetColormap(SimpleConfigAction):

    config_attribute = 'colormap'


class SetRotate(SimpleConfigAction):

    config_attribute = 'rotate'

    def _extra_actions(self, data_info, configuration):
        if self.new_value != self._previous_config.rotate:

            new_rotation = self.new_value - self._previous_config.rotate
            if self._previous_config.flipud:
                new_rotation *= -1

            new_zoom = self._previous_config.zoom.get_rotated(
                new_rotation,
                data_info.get_max_x_index(configuration.dimension, self._previous_config.rotate,
                                          configuration.maps_to_show) + 1,
                data_info.get_max_y_index(configuration.dimension, self._previous_config.rotate,
                                          configuration.maps_to_show) + 1)

            return SetZoom(new_zoom).apply(data_info, configuration)


class SetZoom(SimpleConfigAction):

    config_attribute = 'zoom'


class SetGeneralMask(SimpleConfigAction):

    config_attribute = 'mask_name'


class SetPlotTitle(SimpleConfigAction):

    config_attribute = 'title'


class SetMapTitle(SimpleMapSpecificConfigAction):

    config_attribute = 'title'


class SetMapColorbarLabel(SimpleMapSpecificConfigAction):

    config_attribute = 'colorbar_label'


class SetMapScale(SimpleMapSpecificConfigAction):

    config_attribute = 'scale'


class SetMapClipping(SimpleMapSpecificConfigAction):

    config_attribute = 'clipping'


class SetMapColormap(SimpleMapSpecificConfigAction):

    config_attribute = 'colormap'


class NewConfigAction(ConfigAction):

    def __init__(self, new_config):
        super(NewConfigAction, self).__init__()
        self.new_config = new_config

    def apply(self, data_info, configuration):
        self._previous_config = configuration
        return self.new_config

    def unapply(self):
        return self._previous_config


class SetFont(SimpleConfigAction):

    config_attribute = 'font'


class SetShowAxis(SimpleConfigAction):

    config_attribute = 'show_axis'


class SetColorBarNmrTicks(SimpleConfigAction):

    config_attribute = 'colorbar_nmr_ticks'


class SetInterpolation(SimpleConfigAction):

    config_attribute = 'interpolation'


class SetFlipud(SimpleConfigAction):

    config_attribute = 'flipud'

    def _extra_actions(self, data_info, configuration):
        if self.new_value != self._previous_config.flipud:
            max_y = data_info.get_max_y_index(configuration.dimension, configuration.rotate,
                                              configuration.maps_to_show) + 1

            new_zoom = Zoom(Point(configuration.zoom.p0.x, max_y - configuration.zoom.p1.y),
                            Point(configuration.zoom.p1.x, max_y - configuration.zoom.p0.y))

            return SetZoom(new_zoom).apply(data_info, configuration)



from mdt.models.compartments import CompartmentConfig

__author__ = 'Francisco.Lagos'


class ExpT1DecGRE(CompartmentConfig):

    parameter_list = ('TR', 'flip_angle', 'b1_static', 'T1')
    cl_code = """
        return sin(flip_angle * b1_static) * (1 - exp(-TR / T1)) / (1 - cos(flip_angle * b1_static) * exp(-TR / T1) );
    """


import numpy as np
import pyopencl as cl
from mot.utils import get_float_type_def
from mot.cl_routines.base import CLRoutine
from mot.load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2015-04-16"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DTIMeasures(CLRoutine):

    def concat_and_calculate(self, eigenval1, eigenval2, eigenval3, double_precision=True):
        """Calculate DTI statistics from the given eigenvalues.

        This concatenates the eigenvalue matrices and runs self.calculate(eigenvalues) on them.

        Args:
            eigenval1 (ndarray):
                The first set of eigenvalues, can be 2d, per voxel one eigenvalue,
                or 3d, per voxel multiple eigenvalues.
            eigenval2 (ndarray):
                The first set of eigenvalues, can be 2d, per voxel one eigenvalue,
                or 3d, per voxel multiple eigenvalues.
            eigenval3 (ndarray):
                The first set of eigenvalues, can be 2d, per voxel one eigenvalue,
                or 3d, per voxel multiple eigenvalues.
            double_precision (boolean): if we want to use float (set it to False) or double (set it to True)

        Returns:
            Per voxel, and optionally per instance per voxel, the FA and MD: (fa, md)
        """
        s = eigenval1.shape
        if len(s) < 3:
            return self.calculate(np.concatenate((np.reshape(eigenval1, (s[0], 1)),
                                                  np.reshape(eigenval2, (s[0], 1)),
                                                  np.reshape(eigenval3, (s[0], 1))), axis=1), double_precision)
        else:
            return self.calculate(np.concatenate((np.reshape(eigenval1, (s[0], s[1], 1)),
                                                  np.reshape(eigenval2, (s[0], s[1], 1)),
                                                  np.reshape(eigenval3, (s[0], s[1], 1))), axis=2), double_precision)

    def calculate(self, eigenvalues, double_precision=True):
        """Calculate DTI statistics from the given eigenvalues.

        Args:
            eigenvalues (ndarray):
                The set of eigen values, can be 2d, per voxel one eigenvalue, or 3d, per voxel multiple eigenvalues.
            double_precision (boolean): if we want to use float (set it to False) or double (set it to True)

        Returns:
            Per voxel, and optionally per instance per voxel, the FA and MD: (fa, md)
        """
        np_dtype = np.float32
        if double_precision:
            np_dtype = np.float64

        eigenvalues = np.require(eigenvalues, np_dtype, requirements=['C', 'A', 'O'])

        s = eigenvalues.shape
        if len(s) < 3:
            fa_host = np.zeros((s[0], 1), dtype=np_dtype)
            md_host = np.zeros((s[0], 1), dtype=np_dtype)
            items = s[0]
        else:
            fa_host = np.zeros((s[0] * s[1], 1), dtype=np_dtype)
            md_host = np.zeros((s[0] * s[1], 1), dtype=np_dtype)
            items = s[0] * s[1]
            eigenvalues = np.reshape(eigenvalues, (s[0] * s[1], -1))

        workers = self._create_workers(lambda cl_environment: _DTIMeasuresWorker(
            cl_environment, self.get_compile_flags_list(), eigenvalues, fa_host, md_host, double_precision))
        self.load_balancer.process(workers, items)

        if len(s) > 2:
            fa_host = np.reshape(fa_host, (s[0], s[1], 1))
            md_host = np.reshape(md_host, (s[0], s[1], 1))
        return fa_host, md_host


class _DTIMeasuresWorker(Worker):

    def __init__(self, cl_environment, compile_flags, eigenvalues, fa_host, md_host, double_precision):
        super(_DTIMeasuresWorker, self).__init__(cl_environment)
        self._eigenvalues = eigenvalues
        self._fa_host = fa_host
        self._md_host = md_host
        self._double_precision = double_precision
        self._kernel = self._build_kernel(compile_flags)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        eigenvalues_buf = cl.Buffer(self._cl_run_context.context,
                                    cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                                    hostbuf=self._eigenvalues[range_start:range_end])
        fa_buf = cl.Buffer(self._cl_run_context.context,
                           cl.mem_flags.WRITE_ONLY | cl.mem_flags.COPY_HOST_PTR,
                           hostbuf=self._fa_host[range_start:range_end])
        md_buf = cl.Buffer(self._cl_run_context.context,
                           cl.mem_flags.WRITE_ONLY | cl.mem_flags.COPY_HOST_PTR,
                           hostbuf=self._md_host[range_start:range_end])

        buffers = [eigenvalues_buf, fa_buf, md_buf]

        self._kernel.calculate_measures(self._cl_run_context.queue, (int(nmr_problems), ), None, *buffers)
        cl.enqueue_copy(self._cl_run_context.queue, self._fa_host[range_start:range_end], fa_buf, is_blocking=True)
        event = cl.enqueue_copy(self._cl_run_context.queue, self._md_host[range_start:range_end], md_buf, is_blocking=False)

        return [event]

    def _get_kernel_source(self):
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += '''
            __kernel void calculate_measures(
                global mot_float_type* eigenvalues,
                global mot_float_type* fas,
                global mot_float_type* mds
                ){
                    int gid = get_global_id(0);
                    int voxel = gid * 3;

                    mot_float_type v1 = eigenvalues[voxel];
                    mot_float_type v2 = eigenvalues[voxel + 1];
                    mot_float_type v3 = eigenvalues[voxel + 2];

                    fas[gid] = sqrt(0.5 * (((v1 - v2) * (v1 - v2)) +
                                           ((v1 - v3) * (v1 - v3)) +
                                           ((v2 - v3)  * (v2 - v3))) /
                                                (v1 * v1 + v2 * v2 + v3 * v3));
                    mds[gid] = (v1 + v2 + v3) / 3.0;
            }
        '''
        return kernel_source



from PyQt5.QtCore import QObject, pyqtSignal

from mdt.gui.utils import UpdateDescriptor

__author__ = 'Robbert Harms'
__date__ = "2015-08-20"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SharedState(QObject):

    state_updated_signal = pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        """The shared state for the model fitting GUI

        Attributes:
            base_dir (str): the base dir for all file opening operations
            dimension_index (int): the dimension index used in various operations
            slice_index (int): the slice index used in various operations
        """
        super(SharedState, self).__init__(*args, **kwargs)

        shared_attributes = {'base_dir': None,
                             'dimension_index': 0,
                             'slice_index': 0,
                             'output_folder': None}

        for key, value in shared_attributes.items():
            def get_attribute_setter(attribute_key):
                def setter(value):
                    setattr(self, attribute_key, value)
                return setter

            setattr(self, '_' + key, value)
            setattr(SharedState, key, UpdateDescriptor(key))
            setattr(self, 'set_' + key, get_attribute_setter(key))



import glob
import hashlib
import logging
import os
import shutil
import timeit
from contextlib import contextmanager

import numpy as np
import time
from numpy.lib.format import open_memmap

from mdt.nifti import write_all_as_nifti, get_all_image_data
from mdt.configuration import gzip_optimization_results, gzip_sampling_results
from mdt.utils import create_roi, load_samples

__author__ = 'Robbert Harms'
__date__ = "2016-07-29"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ModelProcessingStrategy(object):

    def __init__(self, tmp_dir=None):
        """Model processing strategies define in how many parts a composite model is processed.

        This uses the :attr:`~mot.model_building.model_builders.OptimizeModelBuilder.problems_to_analyze` attribute
        of the MOT model builder (:class:`mot.model_building.model_builders.OptimizeModelBuilder`) to select the
        voxels to process. That attribute arranges that only a selection of the problems are analyzed
        instead of all of them. We set the range, optimize, collect the data and then repeat until all the voxels
        are optimized.

        Args:
            tmp_dir (str): The temporary dir for the calculations. If set to None we write the temporary results in the
                results folder of each subject. Else, if set set to a specific path we will store the temporary results
                in a subfolder in the given folder (the subfolder will be a hash of the original folder).
        """
        self._logger = logging.getLogger(__name__)
        self._tmp_dir = tmp_dir

    def set_tmp_dir(self, tmp_dir):
        """Set the temporary directory for the calculations. This overwrites the current value.

        Args:
            tmp_dir (str): The temporary dir for the calculations. If set to None we write the temporary results in the
                results folder of each subject. Else, if set to a specific path we will store the temporary results
                in a subfolder in the given folder (the subfolder will be a hash of the original folder).
        """
        self._tmp_dir = tmp_dir
        return self

    def run(self, model, problem_data, output_path, recalculate, worker_generator):
        """Process the given dataset using the logistics the subclass.

        Subclasses of this base class can implement all kind of logic to divide a large dataset in smaller chunks
        (for example slice by slice) and run the processing on each slice separately and join the results afterwards.

        Args:
             model (:class:`~mdt.models.composite.DMRICompositeModel`): the model we want to process
             problem_data (:class:`~mdt.utils.DMRIProblemData`): The problem data object with which
                the model is initialized before running
             output_path (string): The full path to the folder where to place the output
             recalculate (boolean): If we want to recalculate the results if they are already present.
             worker_generator (ModelProcessingWorkerCreator): the generator for creating the worker we will use

        Returns:
            dict: the results as a dictionary of roi lists
        """


class SimpleProcessingStrategy(ModelProcessingStrategy):

    def __init__(self, tmp_dir=None, honor_voxels_to_analyze=True):
        """This class is a base class for all model slice fitting strategies that fit the data in chunks/parts.

        Args:
            honor_voxels_to_analyze (bool): if set to True, we use the model's voxels_to_analyze setting if set
                instead of fitting all voxels in the mask
        """
        super(SimpleProcessingStrategy, self).__init__(tmp_dir=tmp_dir)
        self._honor_voxels_to_analyze = honor_voxels_to_analyze

    @contextmanager
    def _tmp_storage_dir(self, model_output_path, recalculate):
        """Creates a temporary storage dir for the calculations. Removes the dir after calculations.

        Use this manager as a context for running the calculations.

        Args:
            model_output_path (str): the output path of the final model results. We use this to create the tmp_dir.
             recalculate (boolean): if true and the data exists, we throw everything away to start over.
        """
        tmp_storage_dir = self._get_tmp_results_dir(model_output_path)
        self._prepare_tmp_storage_dir(tmp_storage_dir, recalculate)
        yield tmp_storage_dir
        shutil.rmtree(tmp_storage_dir)

    def _get_tmp_results_dir(self, model_output_path):
        """Get the temporary results dir we need to use for processing.

        If self._tmp_dir is set to a non null value we will use a subdirectory in self._tmp_dir.
        Else, if self._tmp_dir is null, we will use a subdir of the model_output_path.

        Args:
            model_output_path (str): the output path of the final model results.

        Returns:
            str: the path to store the temporary results in
        """
        if self._tmp_dir is None:
            return os.path.join(model_output_path, 'tmp_results')

        self._logger.info('Using user defined path for saving the temporary results: {}.'.format(self._tmp_dir))
        return os.path.join(self._tmp_dir, hashlib.md5(model_output_path.encode('utf-8')).hexdigest())

    @staticmethod
    def _prepare_tmp_storage_dir(tmp_storage_dir, recalculate):
        """Prepare the directory for the temporary storage.

        If recalculate is set to True we will remove the storage dir if it exists. Else if False, we will create the
        dir if it does not exist.

        Args:
            tmp_storage_dir (str): the full path to the chunks directory.
            recalculate (boolean): if true and the data exists, we throw everything away to start over.
        """
        if recalculate:
            if os.path.exists(tmp_storage_dir):
                shutil.rmtree(tmp_storage_dir)

        if not os.path.exists(tmp_storage_dir):
            os.makedirs(tmp_storage_dir)

    @contextmanager
    def _selected_indices(self, model, chunk_indices):
        """Create a context in which problems_to_analyze attribute of the models is set to the selected indices.

        Args:
            model: the model to which to set the problems_to_analyze
            chunk_indices (ndarray): the list of voxel indices we want to use for processing
        """
        old_setting = model.problems_to_analyze
        model.problems_to_analyze = chunk_indices
        yield
        model.problems_to_analyze = old_setting


class ChunksProcessingStrategy(SimpleProcessingStrategy):

    def run(self, model, problem_data, output_path, recalculate, worker_generator):
        """Compute all the slices using the implemented chunks generator"""
        start_time = timeit.default_timer()

        with self._tmp_storage_dir(output_path, recalculate) as tmp_storage_dir:
            voxels_processed = 0

            worker = worker_generator.create_worker(model, problem_data, output_path,
                                                    tmp_storage_dir, self._honor_voxels_to_analyze)

            total_roi_indices = worker.get_voxels_to_compute()
            if len(total_roi_indices):
                for chunk_indices in self._chunks_generator(model, problem_data, output_path, worker,
                                                            total_roi_indices):
                    with self._selected_indices(model, chunk_indices):
                        self._run_on_chunk(problem_data, worker, chunk_indices, total_roi_indices,
                                           voxels_processed, start_time)

                    voxels_processed += len(chunk_indices)

            self._logger.info('Computed all voxels, now creating nifti\'s')
            return_data = worker.combine()

        return return_data

    def _chunks_generator(self, model, problem_data, output_path, worker, total_roi_indices):
        """Generate the slices/chunks we will use for the fitting.

        Yields:
            ndarray: the roi indices per chunk we want to process
        """
        raise NotImplementedError

    def _run_on_chunk(self, problem_data, worker, voxel_indices, voxels_to_process, voxels_processed, start_time):
        """Run the worker on the given chunk."""
        total_nmr_voxels = np.count_nonzero(problem_data.mask)
        total_processed = (total_nmr_voxels - len(voxels_to_process)) + voxels_processed

        run_time = timeit.default_timer() - start_time
        current_percentage = total_processed / total_nmr_voxels
        if current_percentage > 0:
            remaining_time = (run_time / current_percentage) - run_time
        else:
            remaining_time = None

        self._logger.info('Computations are at {0:.2%}, processing next {1} voxels ('
                          '{2} voxels in total, {3} processed). Time spent: {4}, time left: {5} (h:m:s).'.
                          format(total_processed / total_nmr_voxels, len(voxel_indices),
                                 total_nmr_voxels,
                                 total_processed,
                                 time.strftime('%H:%M:%S', time.gmtime(run_time)),
                                 time.strftime('%H:%M:%S', time.gmtime(remaining_time)) if remaining_time else '?'))

        worker.process(voxel_indices)


class ModelProcessingWorkerCreator(object):

    def create_worker(self, model, problem_data, output_dir, tmp_storage_dir, honor_voxels_to_analyze):
        """Create and return the worker that the processing strategy will use.

        Args:
            model (:class:`~mdt.models.composite.DMRICompositeModel`): the model we want to process
            problem_data (:class:`~mdt.utils.DMRIProblemData`): The problem data object with which
                the model is initialized before running
            output_dir (str): the location for the final output files
            tmp_storage_dir (str): the location for the temporary output files
            honor_voxels_to_analyze (boolean): if we should honor the voxels_to_analyze list in the model if applicable.

        Returns:
            ModelProcessingWorker: the worker the processing strategy will use.
        """


class SimpleModelProcessingWorkerGenerator(ModelProcessingWorkerCreator):

    def __init__(self, callback_function):
        """Create a generator that will instantiate the worker using the given callback function.

        Args:
            callback_function: the callback function we will call when create_worker is called.
        """
        self._callback_function = callback_function

    def create_worker(self, model, problem_data, output_dir, tmp_storage_dir, honor_voxels_to_analyze):
        return self._callback_function(model, problem_data, output_dir, tmp_storage_dir, honor_voxels_to_analyze)


class ModelProcessingWorker(object):

    def __init__(self, model, problem_data, output_dir, tmp_storage_dir, honor_voxels_to_analyze):
        self._write_volumes_gzipped = True
        self._used_mask_name = 'UsedMask'
        self._model = model
        self._problem_data = problem_data
        self._output_dir = output_dir
        self._tmp_storage_dir = tmp_storage_dir
        self._honor_voxels_to_analyze = honor_voxels_to_analyze
        self._roi_lookup_path = os.path.join(self._tmp_storage_dir, '_roi_voxel_lookup_table.npy')
        self._volume_indices = self._create_roi_to_volume_index_lookup_table()

    def process(self, roi_indices):
        """Process the indicated voxels in the way prescribed by this worker.

        Since the processing strategy can use all voxels to do the analysis in one go, this function
        should return all the output it can, i.e. the same kind of output as from the function :meth:`combine`.

        Args:
            roi_indices (ndarray): The list of roi indices we want to compute

        Returns:
            the results for this single processing step
        """

    def get_voxels_to_compute(self):
        """Get the ROI indices of the voxels we need to compute.

        This should either return an entire list with all the ROI indices for the given brain mask, or a list
        with the specific roi indices we want the strategy to compute.

        Returns:
            ndarray: the list of ROI indices (indexing the current mask) with the voxels we want to compute.
        """
        if self._honor_voxels_to_analyze and self._model.problems_to_analyze:
            roi_list = self._model.problems_to_analyze
        else:
            roi_list = np.arange(0, np.count_nonzero(self._problem_data.mask))

        mask_path = os.path.join(self._tmp_storage_dir, '{}.npy'.format(self._used_mask_name))
        if os.path.exists(mask_path):
            return roi_list[np.logical_not(np.squeeze(create_roi(np.load(mask_path, mmap_mode='r'),
                                                                 self._problem_data.mask)[roi_list]))]
        return roi_list

    def combine(self):
        """Combine all the calculated parts.

        Returns:
            the processing results for as much as this is applicable
        """
        del self._volume_indices
        os.remove(self._roi_lookup_path)

    def _write_volumes(self, roi_indices, results, tmp_dir):
        """Write the result arrays to the temporary storage

        Args:
            roi_indices (ndarray): the indices of the voxels we computed
            results (dict): the dictionary with the results to save
            tmp_dir (str): the directory to save the intermediate results to
        """
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)

        volume_indices = self._volume_indices[roi_indices, :]

        for param_name, result_array in results.items():
            storage_path = os.path.join(tmp_dir, param_name + '.npy')

            map_4d_dim_len = 1
            if len(result_array.shape) > 1:
                map_4d_dim_len = result_array.shape[1]
            else:
                result_array = np.reshape(result_array, (-1, 1))

            mode = 'w+'
            if os.path.isfile(storage_path):
                mode = 'r+'
            tmp_matrix = open_memmap(storage_path, mode=mode, dtype=result_array.dtype,
                                     shape=self._problem_data.mask.shape[0:3] + (map_4d_dim_len,))
            tmp_matrix[volume_indices[:, 0], volume_indices[:, 1], volume_indices[:, 2]] = result_array

        mask_path = os.path.join(tmp_dir, '{}.npy'.format(self._used_mask_name))
        mode = 'w+'
        if os.path.isfile(mask_path):
            mode = 'r+'
        tmp_mask = open_memmap(mask_path, mode=mode, dtype=np.bool, shape=self._problem_data.mask.shape)
        tmp_mask[volume_indices[:, 0], volume_indices[:, 1], volume_indices[:, 2]] = True

    def _combine_volumes(self, output_dir, chunks_dir, volume_header, maps_subdir=''):
        """Combine volumes found in subdirectories to a final volume.

        Args:
            output_dir (str): the location for the output files
            chunks_dir (str): the directory in which all the chunks are located
            maps_subdir (str): we may have per chunk a subdirectory in which the maps are located. This
                parameter is for that subdir. Example search: <chunks_dir>/<chunk>/<maps_subdir>/*.nii*

        Returns:
            dict: the dictionary with the ROIs for every volume, by parameter name
        """
        if not os.path.exists(os.path.join(output_dir, maps_subdir)):
            os.makedirs(os.path.join(output_dir, maps_subdir))

        map_names = list(map(lambda p: os.path.splitext(os.path.basename(p))[0],
                             glob.glob(os.path.join(chunks_dir, maps_subdir, '*.npy'))))

        basic_info = (os.path.join(chunks_dir, maps_subdir),
                      os.path.join(output_dir, maps_subdir),
                      volume_header,
                      self._write_volumes_gzipped)
        info_list = [(map_name, basic_info) for map_name in map_names]

        list(map(_combine_volumes_write_out, info_list))

    def _create_roi_to_volume_index_lookup_table(self):
        """Creates and returns a lookup table for roi index -> volume index.

        This will create from the given mask a memory mapped lookup table mapping the ROI indices (single integer)
        to the correct voxel location in 3d. To find a voxel index using the ROI index, just index this lookup
        table using the ROI index as index.

        For example, suppose we have the lookup table:

            0: (0, 0, 0)
            1: (0, 0, 1)
            2: (0, 1, 0)
            ...

        We can get the position of a voxel in the 3d space by indexing this array as: ``lookup_table[roi_index]``
        to get the correct 3d location.

        Returns:
            memmap ndarray: the memory mapped array which
        """
        if os.path.isfile(self._roi_lookup_path):
            os.remove(self._roi_lookup_path)
        np.save(self._roi_lookup_path, np.argwhere(self._problem_data.mask))
        return np.load(self._roi_lookup_path, mmap_mode='r')


def _combine_volumes_write_out(info_pair):
    """Write out the given information to a nifti volume.

    Needs to be used by ModelProcessingWorker._combine_volumes
    """
    map_name, info_list = info_pair
    chunks_dir, output_dir, volume_header, write_gzipped = info_list

    data = np.load(os.path.join(chunks_dir, map_name + '.npy'), mmap_mode='r')
    write_all_as_nifti({map_name: data}, output_dir, volume_header, gzip=write_gzipped)
    del data


class FittingProcessingWorker(ModelProcessingWorker):

    def __init__(self, optimizer, *args):
        """The processing worker for model fitting.

        Use this if you want to use the model processing strategy to do model fitting.

        Args:
            optimizer: the optimization routine to use
        """
        super(FittingProcessingWorker, self).__init__(*args)
        self._optimizer = optimizer
        self._write_volumes_gzipped = gzip_optimization_results()

    def process(self, roi_indices):
        results, extra_output = self._optimizer.minimize(self._model, full_output=True)
        results.update(extra_output)

        self._write_volumes(roi_indices, results, self._tmp_storage_dir)
        return results

    def combine(self):
        super(FittingProcessingWorker, self).combine()
        self._combine_volumes(self._output_dir, self._tmp_storage_dir, self._problem_data.volume_header)
        return create_roi(get_all_image_data(self._output_dir), self._problem_data.mask)


class SamplingProcessingWorker(ModelProcessingWorker):

    class SampleChainNotStored(object):
        pass

    def __init__(self, sampler, store_samples=False, *args):
        """The processing worker for model sampling.

        Use this if you want to use the model processing strategy to do model sampling.

        Args:
            sampler (AbstractSampler): the optimization sampler to use
            store_samples (boolean): if set to False we will store none of the samples. Use this
                if you are only interested in the volume maps and not in the entire sample chain.
                If set to True the process and combine function will no longer return any results.
        """
        super(SamplingProcessingWorker, self).__init__(*args)
        self._sampler = sampler
        self._write_volumes_gzipped = gzip_sampling_results()
        self._store_samples = store_samples

    def process(self, roi_indices):
        results, other_output = self._sampler.sample(self._model, full_output=True)

        self._write_volumes(roi_indices, other_output, os.path.join(self._tmp_storage_dir, 'volume_maps'))

        if self._store_samples:
            self._write_sample_results(results, self._problem_data.mask, roi_indices)
            return results

        return SamplingProcessingWorker.SampleChainNotStored()

    def combine(self):
        super(SamplingProcessingWorker, self).combine()
        self._combine_volumes(self._output_dir, self._tmp_storage_dir,
                              self._problem_data.volume_header, maps_subdir='volume_maps')

        if self._store_samples:
            for samples in glob.glob(os.path.join(self._tmp_storage_dir, '*.samples.npy')):
                shutil.move(samples, self._output_dir)
            return load_samples(self._output_dir)

        return SamplingProcessingWorker.SampleChainNotStored()

    def _write_sample_results(self, results, full_mask, roi_indices):
        """Write the sample results to a .npy file.

        If the given sample files do not exists or if the existing file is not large enough it will create one
        with enough storage to hold all the samples for the given total_nmr_voxels.
        On storing it should also be given a list of voxel indices with the indices of the voxels that are being stored.

        Args:
            results (dict): the samples to write
            full_mask (ndarray): the complete mask for the entire brain
            roi_indices (ndarray): the roi indices of the voxels we computed
        """
        total_nmr_voxels = np.count_nonzero(full_mask)

        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir)

        for map_name, samples in results.items():
            samples_path = os.path.join(self._output_dir, map_name + '.samples.npy')
            mode = 'w+'
            if os.path.isfile(samples_path):
                mode = 'r+'
                current_results = open_memmap(samples_path, mode='r')
                if current_results.shape[1] != samples.shape[1]:
                    mode = 'w+'
                del current_results # closes the memmap

            saved = open_memmap(samples_path, mode=mode, dtype=samples.dtype,
                                shape=(total_nmr_voxels, samples.shape[1]))
            saved[roi_indices, :] = samples



__author__ = 'Robbert Harms'
__date__ = "2015-10-27"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class DMRIOptimizable(object):

    def __init__(self, *args, **kwargs):
        """This is an interface for some base methods we expect in an MRI model.

        Since we have both composite dMRI models and cascade models we must have an overarching interface to make
        sure that both type of models implement the same additional methods.

        The methods in this interface have little to do with modelling, but unify some extra required methods
        in the cascades and composite models.

        Attributes:
            problems_to_analyze (list): the list with problems we want to analyze. Suppose we have a few thousands
                problems defined in this model, but we want to run the optimization only on a few problems. By setting
                this attribute to a list of problems indices only those problems will be analyzed.
            double_precision (boolean): if we do the computations in double or float precision
        """
        super(DMRIOptimizable, self).__init__()
        self.problems_to_analyze = None
        self.double_precision = False

    def is_protocol_sufficient(self, protocol=None):
        """Check if the protocol holds enough information for this model to work.

        Args:
            protocol (Protocol): The protocol object to check for sufficient information. If set the None, the
                current protocol in the problem data is used.

        Returns:
            boolean: True if there is enough information in the protocol, false otherwise
        """

    def get_protocol_problems(self, protocol=None):
        """Get all the problems with the protocol.

        Args:
            protocol (Protocol): The protocol object to check for problems. If set the None, the
                current protocol in the problem data is used.

        Returns:
            list of ModelProtocolProblem: A list of :class:`~mdt.model_protocol_problem.ModelProtocolProblem` instances
                or subclasses of that baseclass. These objects indicate the problems with the protocol and this model.
        """

    def get_required_protocol_names(self):
        """Get a list with the constant data names that are needed for this model to work.

        For example, an implementing diffusion MRI model might require the presence of the protocol parameter
        ``g`` and ``b``. This function should then return ``('g', 'b')``.

        Returns:
            :class:`list`: A list of columns names that are to be taken from the protocol data.
        """



from mdt.models.compartments import CompartmentConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Dot(CompartmentConfig):

    parameter_list = ()
    cl_code = 'return (mot_float_type)1.0;'



"""Implements multiple handles that hook into the Python logging module.

These handlers can for example echo the log entry to the terminal, write it to a file or dispatch it to another class.
They are typically configured in the MDT configuration file.
"""

import codecs
from logging import StreamHandler
import os
import sys

__author__ = 'Robbert Harms'
__date__ = "2015-08-19"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ModelOutputLogHandler(StreamHandler):

    __instances__ = set()

    def __init__(self, mode='a', encoding=None):
        """This logger logs information about a model optimization to the folder of the model that is being optimized.

        It is by default (see the MDT configuration) already constructed and added to the logging module. To set a new
        file, or to disable this logger set the file using the :attr:`output_file` property.
        """
        super(ModelOutputLogHandler, self).__init__()
        self.__class__.__instances__.add(self)

        if codecs is None:
            encoding = None

        self._output_file = None
        self.mode = mode
        self.encoding = encoding
        self.stream = None

    @property
    def output_file(self):
        return self._output_file

    @output_file.setter
    def output_file(self, output_file):
        self.close()

        self._output_file = output_file
        if self._output_file:
            if not os.path.isdir(os.path.dirname(self._output_file)):
                os.makedirs(os.path.dirname(self._output_file))
            self._open()

    def emit(self, record):
        if self._output_file and self.stream:
            super(ModelOutputLogHandler, self).emit(record)

    def close(self):
        if self._output_file:
            self.acquire()
            try:
                if self.stream:
                    self.flush()
                    if hasattr(self.stream, "close"):
                        self.stream.close()
                    self.stream = None
                super(ModelOutputLogHandler, self).close()
            finally:
                self.release()

    def _open(self):
        """
        Open the current base file with the (original) mode and encoding.
        Return the resulting stream.
        """
        if self._output_file:
            if self.encoding is None:
                self.stream = open(self._output_file, self.mode)
            else:
                self.stream = codecs.open(self._output_file, self.mode, self.encoding)


class StdOutHandler(StreamHandler):

    def __init__(self, stream=None):
        """A redirect for stdout.

        Emits all log entries to the stdout.

        Args:
            stream: the IO stream to which to emit the log entries. If not given we use sys.stdout.
        """
        stream = stream or sys.stdout
        super(StdOutHandler, self).__init__(stream=stream)

    def emit(self, record):
        if self.stream:
            super(StdOutHandler, self).emit(record)


class LogDispatchHandler(StreamHandler):
    _listeners = []

    def __init__(self, *args, **kwargs):
        """This class is able to dispatch messages to all the attached log listeners.

        You can add listeners by adding them to the list of listeners. This list is a class variable and as such is
        available to all instances and subclasses.

        The listeners should be of instance LogListenerInterface.

        This enables for example the GUI to hook a log listener indirectly into the logging module.

        In general only one copy of this class should be used.
        """
        super(LogDispatchHandler, self).__init__(*args, **kwargs)

    def emit(self, record):
        for listener in self._listeners:
            listener.emit(record, self.format(record))

    @staticmethod
    def add_listener(listener):
        """Add a listener to the dispatch handler.

        Args:
            listener (LogListenerInterface): listener that implements the log listener interface.

        Returns:
            int: the listener id number. You can use this to remove the listener again.
        """
        listener_id = len(LogDispatchHandler._listeners)
        LogDispatchHandler._listeners.append(listener)
        return listener_id

    @staticmethod
    def remove_listener(listener_id):
        """Remove a listener from the log dispatcher.

        Args:
            listener_id (int): the id of the listener to remove
        """
        del LogDispatchHandler._listeners[listener_id]


class LogListenerInterface(object):
    """Interface for listeners to work in conjunction with :class:`LogDispatchHandler`"""

    def emit(self, record, formatted_message):
        pass



from mdt.models.composite import DMRICompositeModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class S0(DMRICompositeModelConfig):

    description = 'Models the unweighted signal (aka. b0).'
    model_expression = 'S0'
    volume_selection = {'unweighted_threshold': 25e6,
                        'use_unweighted': True,
                        'use_weighted': False}



from mdt.models.cascade import CascadeConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NODDI(CascadeConfig):

    name = 'NODDI (Cascade)'
    description = 'Cascade for NODDI initialized from Ball&Stick.'
    models = ('BallStick_r1 (Cascade)',
              'NODDI')
    inits = {'NODDI': [('w_ic.w', lambda output_previous, _: output_previous['w_stick.w'] / 2.0),
                       ('w_ec.w', lambda output_previous, _: output_previous['w_stick.w'] / 2.0),
                       ('w_csf.w', 'w_ball.w'),
                       ('NODDI_IC.theta', 'Stick.theta'),
                       ('NODDI_IC.phi', 'Stick.phi')]}


class NODDI_S0(CascadeConfig):

    name = 'NODDI (Cascade|S0)'
    description = 'Cascade for NODDI initialized with only an S0 fit.'
    models = ('S0',
              'NODDI')


class NODDI_Fixed(CascadeConfig):

    name = 'NODDI (Cascade|fixed)'
    description = 'Cascade for NODDI with fixed directions from Ball&Stick.'
    models = ('BallStick_r1 (Cascade)',
              'NODDI')
    inits = {'NODDI': [('w_ic.w', lambda output_previous, _: output_previous['w_stick.w'] / 2.0),
                       ('w_ec.w', lambda output_previous, _: output_previous['w_stick.w'] / 2.0),
                       ('w_csf.w', 'w_ball.w')]}
    fixes = {'NODDI': [('NODDI_IC.theta', 'Stick.theta'),
                       ('NODDI_IC.phi', 'Stick.phi')]}



import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.widgets import Slider
from scipy.stats import norm
import matplotlib.mlab as mlab

__author__ = 'Robbert Harms'
__date__ = "2016-09-02"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SampleVisualizer(object):

    def __init__(self, voxels):
        self._voxels = voxels
        self.voxel_ind = 0
        self.maps_to_show = sorted(self._voxels.keys())
        self.names = {}
        self._figure = None
        self.show_sliders = True
        self._max_voxel_ind = 0
        self._updating_sliders = False
        self._voxel_slider = None
        self._show_trace = True
        self._nmr_bins = 30
        self._show_slider = True
        self._fit_gaussian = True

    def show(self, voxel_ind=0, names=None, maps_to_show=None, to_file=None, block=True, maximize=False,
             show_trace=True, nmr_bins=20, window_title=None, show_sliders=True, fit_gaussian=True,
             figure_options=None):
        """Show the samples per voxel.

        Args:
            voxel_ind (int): the voxel to show the samples from.
            names (dict): A list of names for the different maps. Use as ``{map_name: display_name}`` that is,
                 the key is the name of the map in the volumes dictionary and the display name is the string that will
                 be used as title for that map.
            maps_to_show (:class:`list`): A list of maps to show.
                The items in this list must correspond to the keys in the volumes dictionary.
            to_file (string, optional, default None): If to_file is not None it is supposed
                to be a filename where the image will be saved.
                If not set to None, nothing will be displayed, the results will directly be saved.
                Already existing items will be overwritten.
            block (boolean): If we want to block after calling the plots or not. Set this to False if you
                do not want the routine to block after drawing. In doing so you manually need to block.
            maximize (boolean): if we want to display the window maximized or not
            show_trace (boolean): if we show the trace of each map or not
            nmr_bins (dict or int): either a single value or one per map name
            show_sliders (boolean): if we show the slider or not
            fit_gaussian (boolean): if we fit and show a normal distribution (Gaussian) to the histogram or not
            window_title (str): the title of the window. If None, the default title is used
            figure_options (dict) options for the figure
        """
        figure_options = figure_options or {'figsize': (18, 16)}
        self._figure = plt.figure(**figure_options)

        if names:
            self.names = names
        if maps_to_show:
            self.maps_to_show = maps_to_show
        self.voxel_ind = voxel_ind
        self._nmr_bins = nmr_bins or self._nmr_bins
        self._show_trace = show_trace
        self.show_sliders = show_sliders
        self._fit_gaussian = fit_gaussian

        self._setup()

        if maximize:
            mng = plt.get_current_fig_manager()
            mng.window.showMaximized()

        if window_title:
            mng = plt.get_current_fig_manager()
            mng.canvas.set_window_title(window_title)

        if to_file:
            plt.savefig(to_file)
            plt.close()
        else:
            plt.draw()
            if block:
                plt.show(True)

    def set_voxel(self, voxel_ind):
        voxel_ind = round(voxel_ind)
        if not self._updating_sliders:
            self._updating_sliders = True
            self.voxel_ind = int(round(voxel_ind))

            self._voxel_slider.set_val(voxel_ind)
            self._rerender()
            self._voxel_slider.set_val(voxel_ind)
            self._updating_sliders = False

    def _setup(self):
        self._rerender()

        self._max_voxel_ind = max([self._voxels[map_name].shape[0] for map_name in self.maps_to_show])

        y_positions = [0.008]

        if self.show_sliders:
            ax = self._figure.add_axes([0.25, y_positions[0], 0.5, 0.01], axisbg='Wheat')
            self._voxel_slider = _DiscreteSlider(ax, 'Voxel', 0, self._max_voxel_ind - 1,
                                                     valinit=self.voxel_ind, valfmt='%i',
                                                     color='DarkSeaGreen', closedmin=True, closedmax=True)
            self._voxel_slider.on_changed(self.set_voxel)

    def _rerender(self):
        nmr_maps = len(self.maps_to_show)
        if self._show_trace:
            nmr_maps *= 2

        grid = GridSpec(nmr_maps, 1, left=0.04, right=0.96, top=0.94, bottom=0.06, hspace=0.2)

        i = 0
        for map_name in self.maps_to_show:
            samples = self._voxels[map_name]

            title = map_name
            if map_name in self.names:
                title = self.names[map_name]

            if isinstance(self._nmr_bins, dict) and map_name in self._nmr_bins:
                nmr_bins = self._nmr_bins[map_name]
            else:
                nmr_bins = self._nmr_bins

            hist_plot = plt.subplot(grid[i])
            n, bins, patches = hist_plot.hist(samples[self.voxel_ind, :], nmr_bins, normed=True)
            plt.title(title)
            i += 1

            if self._fit_gaussian:
                mu, sigma = norm.fit(samples[self.voxel_ind, :])
                bincenters = 0.5*(bins[1:] + bins[:-1])
                y = mlab.normpdf(bincenters, mu, sigma)
                hist_plot.plot(bincenters, y, 'r', linewidth=1)

            if self._show_trace:
                trace_plot = plt.subplot(grid[i])
                trace_plot.plot(samples[self.voxel_ind, :])
                i += 1


class _DiscreteSlider(Slider):
    """A matplotlib slider widget with discrete steps."""

    def __init__(self, *args, **kwargs):
        """Identical to Slider.__init__, except for the "increment" and kwarg.

        Args:
            increment (float): specifies the step size that the slider will be discritized
        to."""
        self.inc = kwargs.pop('increment', 0.25)
        Slider.__init__(self, *args, **kwargs)

    def set_max(self, new_max):
        orig_val = self.val
        self.set_val(self.valmin)

        self.valmax = new_max
        self.ax.set_xlim((self.valmin, self.valmax))

        if orig_val >= new_max:
            self.set_val((new_max + self.valmin) / 2.0)
        else:
            self.set_val(orig_val)

    def set_val(self, val):
        discrete_val = int(val / self.inc) * self.inc
        # We can't just call Slider.set_val(self, discrete_val), because this
        # will prevent the slider from updating properly (it will get stuck at
        # the first step and not "slide"). Instead, we'll keep track of the
        # the continuous value as self.val and pass in the discrete value to
        # everything else.
        xy = self.poly.xy
        xy[2] = discrete_val, 1
        xy[3] = discrete_val, 0
        self.poly.xy = xy
        self.valtext.set_text(self.valfmt % discrete_val)
        if self.drawon:
            self.ax.figure.canvas.draw()
        self.val = val
        if not self.eventson:
            return
        for cid, func in self.observers.items():
            func(discrete_val)



"""Definitions of the protocol parameters

The type of these parameters signifies that the data for this parameter should come from the protocol defined in the
model data. These will never be optimized and are always set to the data defined in the protocol.

Please choose the parameter type for a model and parameter carefully since the type signifies how the parameter and
its data are handled during model construction.

"""
from mdt.models.parameters import ProtocolParameterConfig

__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class g(ProtocolParameterConfig):
    data_type = 'mot_float_type4'


class b(ProtocolParameterConfig):
    pass


class G(ProtocolParameterConfig):
    pass


class Delta(ProtocolParameterConfig):
    pass


class delta(ProtocolParameterConfig):
    pass


class TE(ProtocolParameterConfig):
    pass


class TM(ProtocolParameterConfig):
    pass


class Ti(ProtocolParameterConfig):
    pass


class TR(ProtocolParameterConfig):
    pass


class flip_angle(ProtocolParameterConfig):
    pass

# For STEAM/TSE sequences, depending on the model in which they are used.
class Refoc_fa1(ProtocolParameterConfig):
    pass


# For STEAM/TSE sequences, depending on the model in which they are used.
class Refoc_fa2(ProtocolParameterConfig):
    pass


# For STEAM/TSE sequences, depending on the model in which they are used.
class SEf(ProtocolParameterConfig):
    pass



from mdt.models.compartments import CompartmentConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ExpT2Dec(CompartmentConfig):

    parameter_list = ('TE', 'T2')
    cl_code = 'return exp(-TE / T2);'



import datetime
import hashlib
import os

from six import string_types

__author__ = 'Robbert Harms'
__date__ = "2016-08-04"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SaveUserScriptInfo(object):

    def __init__(self):
        """Base class for writing the user info to a file.
        """

    def write(self, output_file):
        """Write the information about the script the user is executing to a output information file.

        Args:
            output_file (str): the file we will write the output to
        """
        raise NotImplementedError


def easy_save_user_script_info(save_user_script_info, output_file, file_path_from_stack):
    """Handy routine for saving the user script info from multiple sources.

    Args:
        save_user_script_info (boolean, str or SaveUserScriptInfo): The info we need to save about the script the
            user is currently executing. If True (default) we use the stack to lookup the script the user is executing
            and save that using a SaveFromScript saver. If a string is given we use that filename again for the
            SaveFromScript saver. If False or None, we do not write any information. If a SaveUserScriptInfo is
            given we use that directly.
        output_file (str): the output folder for the output file
        file_path_from_stack (str): the file path from the stack inspection
    """
    if save_user_script_info:
        if isinstance(save_user_script_info, SaveUserScriptInfo):
            save_user_script_info.write(output_file)

        elif isinstance(save_user_script_info, string_types):
            SaveFromScript(save_user_script_info).write(output_file)

        elif save_user_script_info is True:
            SaveFromScript(file_path_from_stack).write(output_file)


class SaveFromScript(SaveUserScriptInfo):

    def __init__(self, user_script_path):
        super(SaveFromScript, self).__init__()
        self._user_script_path = user_script_path

    def write(self, output_file):
        """Write the information about the script the user is executing to a output information file.

        This function relies on the caller of the script to provide the filename of the script he user is currently
        executing. You can do this using for example a stack lookup:

        .. code-block:: python

            stack()[1][0].f_globals.get('__file__')

        Args:
            output_file (str): the file we will write the output to
        """
        output_file_existed = os.path.exists(output_file)

        utc_time = datetime.datetime.utcnow()

        class_hash_name = hashlib.sha1()
        class_hash_name.update(str(utc_time).encode('utf-8'))

        with open(output_file, 'a') as output_file:
            if not output_file_existed:
                output_file.write('from mdt.user_script_info import UserScriptInfo')

            output_file.write('\n\n\nclass Script_{}(UserScriptInfo):\n\n'.format(class_hash_name.hexdigest()))
            output_file.write(' '*4 + 'date = "{}"\n'.format(utc_time))
            output_file.write(' '*4 + 'filename = "{}"\n\n'.format(self._user_script_path))
            output_file.write(' ' * 4 + '@staticmethod\n'.format(self._user_script_path))
            output_file.write(' ' * 4 + 'def body():\n'.format(self._user_script_path))

            with open(self._user_script_path, 'r') as input_file:
                for line in input_file:
                    output_file.write(' '*8 + line)


class UserScriptInfo(object):
    pass
